# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import List
from unittest import mock

from IPython.testing import globalipapp
import pytest


@pytest.mark.parametrize(
    ("sql_template", "expected"),
    (
        (
            "{my_project}.{my_dataset}.{my_table}",
            ["my_project", "my_dataset", "my_table"],
        ),
        (
            "{{not a format variable}}",
            [],
        ),
    ),
)
def test_parse_fields(sql_template: str, expected: List[str]):
    import bigquery_magics.pyformat

    fields = bigquery_magics.pyformat._parse_fields(sql_template)
    fields.sort()
    expected.sort()
    assert fields == expected


@pytest.mark.usefixtures("mock_credentials")
def test_pyformat_with_unsupported_type_raises_typeerror(ipython_ns_cleanup):
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")
    ipython_ns_cleanup.extend(
        [
            (ip, "my_object"),
        ]
    )
    ip.user_ns["my_object"] = object()

    sql = "SELECT {my_object}"

    with pytest.raises(TypeError, match="my_object has unsupported type: "):
        ip.run_cell_magic("bigquery", "--pyformat", sql)


@pytest.mark.usefixtures("mock_credentials")
def test_pyformat_with_missing_variable_raises_keyerror():
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")

    sql = "SELECT {my_object}"

    with pytest.raises(KeyError, match="my_object"):
        ip.run_cell_magic("bigquery", "--pyformat", sql)


@pytest.mark.usefixtures("mock_credentials")
def test_pyformat_with_no_variables():
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")

    run_query_patch = mock.patch("bigquery_magics.bigquery._run_query", autospec=True)

    sql = "SELECT '{{escaped curly brackets}}'"
    expected_sql = "SELECT '{escaped curly brackets}'"

    with run_query_patch as run_query_mock:
        ip.run_cell_magic("bigquery", "--pyformat", sql)
        run_query_mock.assert_called_once_with(mock.ANY, expected_sql, mock.ANY)


@pytest.mark.usefixtures("mock_credentials")
def test_pyformat_with_query_string_replaces_variables(ipython_ns_cleanup):
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")
    ipython_ns_cleanup.extend(
        [
            # String
            (ip, "project"),
            (ip, "dataset"),
            (ip, "table"),
            (ip, "my_string"),
            # Float
            (ip, "max_value"),
            # Integer
            (ip, "year"),
            # Unused, unsupported types shouldn't raise.
            (ip, "my_object"),
        ]
    )
    ip.user_ns["project"] = "pyformat-project"
    ip.user_ns["dataset"] = "pyformat_dataset"
    ip.user_ns["table"] = "pyformat_table"
    ip.user_ns["my_string"] = "some string value"
    ip.user_ns["max_value"] = 2.25
    ip.user_ns["year"] = 2025
    ip.user_ns["my_object"] = object()

    sql = """
    SELECT {year} - year  AS age,
    @myparam AS myparam,
    '{{my_string}}' AS escaped_string,
    FROM `{project}.{dataset}.{table}`
    WHERE height < {max_value}
    """
    expected_sql = """
    SELECT 2025 - year  AS age,
    @myparam AS myparam,
    '{my_string}' AS escaped_string,
    FROM `pyformat-project.pyformat_dataset.pyformat_table`
    WHERE height < 2.25
    """.strip()

    run_query_patch = mock.patch("bigquery_magics.bigquery._run_query", autospec=True)

    with run_query_patch as run_query_mock:
        ip.run_cell_magic("bigquery", "--pyformat --params {'myparam': 42}", sql)
        run_query_mock.assert_called_once_with(mock.ANY, expected_sql, mock.ANY)


@pytest.mark.usefixtures("mock_credentials")
def test_pyformat_with_query_dollar_variable_replaces_variables(ipython_ns_cleanup):
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")

    ipython_ns_cleanup.extend(
        [
            (ip, "custom_query"),
            (ip, "my_string"),
        ]
    )
    run_query_patch = mock.patch("bigquery_magics.bigquery._run_query", autospec=True)

    sql = "SELECT 42, '{my_string}'"
    expected_sql = "SELECT 42, 'This is a test'"
    ip.user_ns["my_string"] = "This is a test"
    ip.user_ns["custom_query"] = sql

    cell_body = "$custom_query"  # Referring to an existing variable name (custom_query)

    with run_query_patch as run_query_mock:
        ip.run_cell_magic("bigquery", "--pyformat", cell_body)
        run_query_mock.assert_called_once_with(mock.ANY, expected_sql, mock.ANY)


@pytest.mark.usefixtures("mock_credentials")
def test_without_pyformat_doesnt_modify_curly_brackets(ipython_ns_cleanup):
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    ip.extension_manager.load_extension("bigquery_magics")

    ipython_ns_cleanup.extend(
        [
            (ip, "my_string"),
        ]
    )
    run_query_patch = mock.patch("bigquery_magics.bigquery._run_query", autospec=True)

    sql = "SELECT 42, '{my_string}'"
    expected_sql = "SELECT 42, '{my_string}'"  # No pyformat means no escaping needed.
    ip.user_ns["my_string"] = "This is a test"

    with run_query_patch as run_query_mock:
        ip.run_cell_magic("bigquery", "", sql)
        run_query_mock.assert_called_once_with(mock.ANY, expected_sql, mock.ANY)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

import pytest
import requests

try:
    import spanner_graphs.graph_visualization as graph_visualization
except ImportError:
    graph_visualization = None

import bigquery_magics.graph_server as graph_server

alex_properties = {
    "birthday": "1991-12-21T08:00:00Z",
    "id": 1,
    "city": "Adelaide",
    "country": "Australia",
    "name": "Alex",
}

alex_account_properties = {
    "create_time": "2020-01-10T14:22:20.222Z",
    "id": 7,
    "is_blocked": False,
    "nick_name": "Vacation Fund",
}

alex_owns_account_edge_properites = {
    "account_id": 7,
    "create_time": "2020-01-10T14:22:20.222Z",
    "id": 1,
}

row_alex_owns_account = [
    {
        "identifier": "mUZpbkdyYXBoLlBlcnNvbgB4kQI=",
        "kind": "node",
        "labels": ["Person"],
        "properties": alex_properties,
    },
    {
        "destination_node_identifier": "mUZpbkdyYXBoLkFjY291bnQAeJEO",
        "identifier": "mUZpbkdyYXBoLlBlcnNvbk93bkFjY291bnQAeJECkQ6ZRmluR3JhcGguUGVyc29uAHiRAplGaW5HcmFwaC5BY2NvdW50AHiRDg==",
        "kind": "edge",
        "labels": ["Owns"],
        "properties": alex_owns_account_edge_properites,
        "source_node_identifier": "mUZpbkdyYXBoLlBlcnNvbgB4kQI=",
    },
    {
        "identifier": "mUZpbkdyYXBoLkFjY291bnQAeJEO",
        "kind": "node",
        "labels": ["Account"],
        "properties": alex_account_properties,
    },
]

lee_properties = {
    "birthday": "1986-12-07T08:00:00Z",
    "city": "Kollam",
    "country": "India",
    "id": 3,
    "name": "Lee",
}

lee_account_properties = {
    "create_time": "2020-01-28T01:55:09.206Z",
    "id": 16,
    "is_blocked": True,
    "nick_name": "Vacation Fund",
}

lee_owns_account_edge_properties = {
    "account_id": 16,
    "create_time": "2020-02-18T13:44:20.655Z",
    "id": 3,
}

row_lee_owns_account = [
    {
        "identifier": "mUZpbkdyYXBoLlBlcnNvbgB4kQY=",
        "kind": "node",
        "labels": ["Person"],
        "properties": lee_properties,
    },
    {
        "destination_node_identifier": "mUZpbkdyYXBoLkFjY291bnQAeJEg",
        "identifier": "mUZpbkdyYXBoLlBlcnNvbk93bkFjY291bnQAeJEGkSCZRmluR3JhcGguUGVyc29uAHiRBplGaW5HcmFwaC5BY2NvdW50AHiRIA==",
        "kind": "edge",
        "labels": ["Owns"],
        "properties": lee_owns_account_edge_properties,
        "source_node_identifier": "mUZpbkdyYXBoLlBlcnNvbgB4kQY=",
    },
    {
        "identifier": "mUZpbkdyYXBoLkFjY291bnQAeJEg",
        "kind": "node",
        "labels": ["Account"],
        "properties": lee_account_properties,
    },
]


def _validate_nodes_and_edges(result):
    for edge in result["response"]["edges"]:
        assert "source_node_identifier" in edge
        assert "destination_node_identifier" in edge
        assert "identifier" in edge
        assert "Owns" in edge["labels"]
        assert "properties" in edge

    print(result["response"]["nodes"])
    for node in result["response"]["nodes"]:
        assert "identifier" in node
        assert "Account" in node["labels"] or "Person" in node["labels"]
        assert "properties" in node


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_one_column_no_rows():
    result = graph_server.convert_graph_data({"result": {}})
    assert result == {
        "response": {
            "edges": [],
            "nodes": [],
            "query_result": {"result": []},
            "schema": None,
        }
    }


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_one_column_one_row_one_column():
    result = graph_server.convert_graph_data(
        {
            "result": {
                "0": json.dumps(row_alex_owns_account),
            }
        }
    )

    assert len(result["response"]["nodes"]) == 2
    assert len(result["response"]["edges"]) == 1

    _validate_nodes_and_edges(result)

    assert result["response"]["query_result"] == {"result": [row_alex_owns_account]}
    assert result["response"]["schema"] is None


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_one_column_two_rows_one_column_null_json():
    result = graph_server.convert_graph_data(
        {
            "result": {
                "0": json.dumps(None),
                "1": json.dumps(row_alex_owns_account),
            }
        }
    )

    # Null JSON element should be ignored in visualization, but should still be present in tabular view.
    assert len(result["response"]["nodes"]) == 2
    assert len(result["response"]["edges"]) == 1

    _validate_nodes_and_edges(result)

    assert result["response"]["query_result"] == {
        "result": [None, row_alex_owns_account]
    }
    assert result["response"]["schema"] is None

    _validate_nodes_and_edges(result)


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_one_column_two_rows():
    result = graph_server.convert_graph_data(
        {
            "result": {
                "0": json.dumps(row_alex_owns_account),
                "1": json.dumps(row_lee_owns_account),
            }
        }
    )

    assert len(result["response"]["nodes"]) == 4
    assert len(result["response"]["edges"]) == 2

    _validate_nodes_and_edges(result)

    assert result["response"]["query_result"] == {
        "result": [row_alex_owns_account, row_lee_owns_account]
    }
    assert result["response"]["schema"] is None


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_one_row_two_columns():
    result = graph_server.convert_graph_data(
        {
            "col1": {
                "0": json.dumps(row_alex_owns_account),
            },
            "col2": {
                "0": json.dumps(row_lee_owns_account),
            },
        }
    )

    assert len(result["response"]["nodes"]) == 4
    assert len(result["response"]["edges"]) == 2

    _validate_nodes_and_edges(result)

    assert result["response"]["query_result"] == {
        "col1": [row_alex_owns_account],
        "col2": [row_lee_owns_account],
    }
    assert result["response"]["schema"] is None


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_nongraph_json():
    # If we have valid json that doesn't represent a graph, we don't expect to get nodes and edges,
    # but we should at least have row data, allowing the tabular view to work.
    result = graph_server.convert_graph_data(
        {
            "result": {
                "0": json.dumps({"foo": 1, "bar": 2}),
            }
        }
    )

    assert len(result["response"]["nodes"]) == 0
    assert len(result["response"]["edges"]) == 0

    assert result["response"]["query_result"] == {"result": [{"foo": 1, "bar": 2}]}
    assert result["response"]["schema"] is None


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_outer_key_not_string():
    result = graph_server.convert_graph_data(
        {
            0: {
                "0": json.dumps({"foo": 1, "bar": 2}),
            }
        }
    )
    assert result == {"error": "Expected outer key to be str, got <class 'int'>"}


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_outer_value_not_dict():
    result = graph_server.convert_graph_data({"result": 0})
    assert result == {"error": "Expected outer value to be dict, got <class 'int'>"}


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_inner_value_not_string():
    result = graph_server.convert_graph_data(
        {
            "col1": {
                "0": json.dumps(row_alex_owns_account),
            },
            "col2": {
                "0": 12345,
            },
        }
    )

    # Non-JSON column should be ignored in visualizer view, but still appear in tabular view.
    assert len(result["response"]["nodes"]) == 2
    assert len(result["response"]["edges"]) == 1

    _validate_nodes_and_edges(result)

    assert result["response"]["query_result"] == {
        "col1": [row_alex_owns_account],
        "col2": ["12345"],
    }
    assert result["response"]["schema"] is None


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_empty_dict():
    result = graph_server.convert_graph_data({})
    assert result == {
        "response": {
            "nodes": [],
            "edges": [],
            "schema": None,
            "query_result": {},
        }
    }


@pytest.mark.skipif(
    graph_visualization is None, reason="Requires `spanner-graph-notebook`"
)
def test_convert_wrong_row_index():
    result0 = graph_server.convert_graph_data(
        {
            "result": {
                "0": json.dumps(row_alex_owns_account),
            }
        }
    )

    # Changing the index should not impact the result.
    result1 = graph_server.convert_graph_data(
        {
            "result": {
                "1": json.dumps(row_alex_owns_account),
            }
        }
    )

    assert result1 == result0


class TestGraphServer(unittest.TestCase):
    def setUp(self):
        if graph_visualization is not None:  # pragma: NO COVER
            self.server_thread = graph_server.graph_server.init()

    def tearDown(self):
        if graph_visualization is not None:  # pragma: NO COVER
            graph_server.graph_server.stop_server()  # Stop the server after each test
            self.server_thread.join()  # Wait for the thread to finish

    @pytest.mark.skipif(
        graph_visualization is None, reason="Requires `spanner-graph-notebook`"
    )
    def test_get_ping(self):
        self.assertTrue(self.server_thread.is_alive())

        route = graph_server.graph_server.build_route(
            graph_server.GraphServer.endpoints["get_ping"]
        )
        response = requests.get(route)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"message": "pong"})

    @pytest.mark.skipif(
        graph_visualization is None, reason="Requires `spanner-graph-notebook`"
    )
    def test_post_ping(self):
        self.assertTrue(self.server_thread.is_alive())
        route = graph_server.graph_server.build_route(
            graph_server.GraphServer.endpoints["post_ping"]
        )
        response = requests.post(route, json={"data": "ping"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"your_request": {"data": "ping"}})

    @pytest.mark.skipif(
        graph_visualization is None, reason="Requires `spanner-graph-notebook`"
    )
    def test_post_query(self):
        self.assertTrue(self.server_thread.is_alive())
        route = graph_server.graph_server.build_route(
            graph_server.GraphServer.endpoints["post_query"]
        )

        data = {
            "result": {
                "0": json.dumps(row_alex_owns_account),
            }
        }
        response = requests.post(route, json={"params": json.dumps(data)})
        self.assertEqual(response.status_code, 200)
        response_data = response.json()["response"]

        self.assertEqual(len(response_data["nodes"]), 2)
        self.assertEqual(len(response_data["edges"]), 1)

        _validate_nodes_and_edges(response.json())

        self.assertEqual(
            response_data["query_result"], {"result": [row_alex_owns_account]}
        )
        self.assertIsNone(response_data["schema"])

    @pytest.mark.skipif(
        graph_visualization is None, reason="Requires `spanner-graph-notebook`"
    )
    def test_post_node_expansion(self):
        self.assertTrue(self.server_thread.is_alive())
        route = graph_server.graph_server.build_route(
            graph_server.GraphServer.endpoints["post_node_expansion"]
        )
        request = {
            "request": {
                "uid": "test_uid",
                "node_labels": ["label1, label2"],
                "node_properites": {},
                "direction": "INCOMING",
                "edge_label": None,
            },
            "params": "{}",
        }
        response = requests.post(route, json={"params": json.dumps(request)})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(), {"error": "Node expansion not yet implemented"}
        )

    @pytest.mark.skipif(
        graph_visualization is None, reason="Requires `spanner-graph-notebook`"
    )
    def test_post_node_expansion_invalid_request(self):
        self.assertTrue(self.server_thread.is_alive())
        route = graph_server.graph_server.build_route(
            graph_server.GraphServer.endpoints["post_node_expansion"]
        )
        request = {}
        response = requests.post(route, json={"params": json.dumps(request)})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(), {"error": "Node expansion not yet implemented"}
        )


def test_stop_server_never_started():
    graph_server.graph_server.stop_server()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bigquery_magics.line_arg_parser.exceptions as lap_exceptions
import bigquery_magics.line_arg_parser.lexer as lap_lexer


class ParseNode(object):
    """A base class for nodes in the input parsed to an abstract syntax tree."""


class InputLine(ParseNode):
    def __init__(self, destination_var, option_list):
        self.destination_var = destination_var
        self.option_list = option_list


class DestinationVar(ParseNode):
    def __init__(self, token):
        # token type is DEST_VAR
        self.token = token
        self.name = token.lexeme if token is not None else None


class CmdOptionList(ParseNode):
    def __init__(self, option_nodes):
        self.options = [node for node in option_nodes]  # shallow copy


class CmdOption(ParseNode):
    def __init__(self, name, value):
        self.name = name  # string
        self.value = value  # CmdOptionValue node


class ParamsOption(CmdOption):
    def __init__(self, value):
        super(ParamsOption, self).__init__("params", value)


class CmdOptionValue(ParseNode):
    def __init__(self, token):
        # token type is OPT_VAL
        self.token = token
        self.value = token.lexeme


class PyVarExpansion(ParseNode):
    def __init__(self, token):
        self.token = token
        self.raw_value = token.lexeme


class PyDict(ParseNode):
    def __init__(self, dict_items):
        self.items = [item for item in dict_items]  # shallow copy


class PyDictItem(ParseNode):
    def __init__(self, key, value):
        self.key = key
        self.value = value


class PyDictKey(ParseNode):
    def __init__(self, token):
        self.token = token
        self.key_value = token.lexeme


class PyScalarValue(ParseNode):
    def __init__(self, token, raw_value):
        self.token = token
        self.raw_value = raw_value


class PyTuple(ParseNode):
    def __init__(self, tuple_items):
        self.items = [item for item in tuple_items]  # shallow copy


class PyList(ParseNode):
    def __init__(self, list_items):
        self.items = [item for item in list_items]  # shallow copy


class Parser(object):
    """Parser for the tokenized cell magic input line.

    The parser recognizes a simplified subset of Python grammar, specifically
    a dictionary representation in typical use cases when the "--params" option
    is used with the %%bigquery cell magic.

    The grammar (terminal symbols are CAPITALIZED):

        input_line       : destination_var option_list
        destination_var  : DEST_VAR | EMPTY
        option_list      : (OPTION_SPEC [OPTION_EQ] option_value)*
                           (params_option | EMPTY)
                           (OPTION_SPEC [OPTION_EQ] option_value)*

        option_value     : OPT_VAL | EMPTY

        # DOLLAR_PY_ID can occur if a variable passed to --params does not exist
        # and is thus not expanded to a dict.
        params_option    : PARAMS_OPT_SPEC [PARAMS_OPT_EQ] \
                           (DOLLAR_PY_ID | PY_STRING | py_dict)

        py_dict          : LCURL dict_items RCURL
        dict_items       : dict_item | (dict_item COMMA dict_items)
        dict_item        : (dict_key COLON py_value) | EMPTY

        # dict items are actually @parameter names in the cell body (i.e. the query),
        # thus restricting them to strings.
        dict_key         : PY_STRING

        py_value         : PY_BOOL
                         | PY_NUMBER
                         | PY_STRING
                         | py_tuple
                         | py_list
                         | py_dict

        py_tuple         : LPAREN collection_items RPAREN
        py_list          : LSQUARE collection_items RSQUARE
        collection_items : collection_item | (collection_item COMMA collection_items)
        collection_item  : py_value | EMPTY

    Args:
        lexer (line_arg_parser.lexer.Lexer):
            An iterable producing a tokenized cell magic argument line.
    """

    def __init__(self, lexer):
        self._lexer = lexer
        self._tokens_iter = iter(self._lexer)
        self.get_next_token()

    def get_next_token(self):
        """Obtain the next token from the token stream and store it as current."""
        token = next(self._tokens_iter)
        self._current_token = token

    def consume(self, expected_type, exc_type=lap_exceptions.ParseError):
        """Move to the next token in token stream if it matches the expected type.

        Args:
            expected_type (bigquery_magics.line_arg_parser.lexer.TokenType): The expected token type to be consumed.
            exc_type (Optional[ParseError]): The type of the exception to raise. Should be
                the ``ParseError`` class or one of its subclasses. Defaults to
                ``ParseError``.

        Raises:
            ParseError: If the current token does not match the expected type.
        """
        if self._current_token.type_ == expected_type:
            if expected_type != lap_lexer.TokenType.EOL:
                self.get_next_token()
        else:
            if self._current_token.type_ == lap_lexer.TokenType.EOL:
                msg = "Unexpected end of input, expected {}.".format(expected_type)
            else:
                msg = "Expected token type {}, but found {} at position {}.".format(
                    expected_type, self._current_token.lexeme, self._current_token.pos
                )
            self.error(message=msg, exc_type=exc_type)

    def error(self, message="Syntax error.", exc_type=lap_exceptions.ParseError):
        """Raise an error with the given message.

        Args:
            expected_type (bigquery_magics.line_arg_parser.lexer.TokenType): The expected token type to be consumed.
            exc_type (Optional[ParseError]): The type of the exception to raise. Should be
                the ``ParseError`` class or one of its subclasses. Defaults to
                ``ParseError``.

        Raises:
            ParseError: If the current token does not match the expected type.
        """
        raise exc_type(message)

    def input_line(self):
        """The top level method for parsing the cell magic arguments line.

        Implements the following grammar production rule:

            input_line : destination_var option_list
        """
        dest_var = self.destination_var()
        options = self.option_list()

        token = self._current_token

        if token.type_ != lap_lexer.TokenType.EOL:
            msg = "Unexpected input at position {}: {}".format(token.pos, token.lexeme)
            self.error(msg)

        return InputLine(dest_var, options)

    def destination_var(self):
        """Implementation of the ``destination_var`` grammar production rule.

        Production:

            destination_var  : DEST_VAR | EMPTY
        """
        token = self._current_token

        if token.type_ == lap_lexer.TokenType.DEST_VAR:
            self.consume(lap_lexer.TokenType.DEST_VAR)
            result = DestinationVar(token)
        elif token.type_ == lap_lexer.TokenType.UNKNOWN:
            msg = "Unknown input at position {}: {}".format(token.pos, token.lexeme)
            self.error(msg)
        else:
            result = DestinationVar(None)

        return result

    def option_list(self):
        """Implementation of the ``option_list`` grammar production rule.

        Production:

            option_list : (OPTION_SPEC [OPTION_EQ] option_value)*
                          (params_option | EMPTY)
                          (OPTION_SPEC [OPTION_EQ] option_value)*
        """
        all_options = []

        def parse_nonparams_options():
            while self._current_token.type_ == lap_lexer.TokenType.OPTION_SPEC:
                token = self._current_token
                self.consume(lap_lexer.TokenType.OPTION_SPEC)

                opt_name = token.lexeme[2:]  # cut off the "--" prefix

                # skip the optional "=" character
                if self._current_token.type_ == lap_lexer.TokenType.OPTION_EQ:
                    self.consume(lap_lexer.TokenType.OPTION_EQ)

                opt_value = self.option_value()
                option = CmdOption(opt_name, opt_value)
                all_options.append(option)

        parse_nonparams_options()

        token = self._current_token

        if token.type_ == lap_lexer.TokenType.PARAMS_OPT_SPEC:
            option = self.params_option()
            all_options.append(option)

        parse_nonparams_options()

        if self._current_token.type_ == lap_lexer.TokenType.PARAMS_OPT_SPEC:
            self.error(
                message="Duplicate --params option",
                exc_type=lap_exceptions.DuplicateQueryParamsError,
            )

        return CmdOptionList(all_options)

    def option_value(self):
        """Implementation of the ``option_value`` grammar production rule.

        Production:

            option_value : OPT_VAL | EMPTY
        """
        token = self._current_token

        if token.type_ == lap_lexer.TokenType.OPT_VAL:
            self.consume(lap_lexer.TokenType.OPT_VAL)
            result = CmdOptionValue(token)
        elif token.type_ == lap_lexer.TokenType.UNKNOWN:
            msg = "Unknown input at position {}: {}".format(token.pos, token.lexeme)
            self.error(msg)
        else:
            result = None

        return result

    def params_option(self):
        """Implementation of the ``params_option`` grammar production rule.

        Production:

            params_option : PARAMS_OPT_SPEC [PARAMS_OPT_EQ] \
                            (DOLLAR_PY_ID | PY_STRING | py_dict)
        """
        self.consume(lap_lexer.TokenType.PARAMS_OPT_SPEC)

        # skip the optional "=" character
        if self._current_token.type_ == lap_lexer.TokenType.PARAMS_OPT_EQ:
            self.consume(lap_lexer.TokenType.PARAMS_OPT_EQ)

        if self._current_token.type_ == lap_lexer.TokenType.DOLLAR_PY_ID:
            token = self._current_token
            self.consume(lap_lexer.TokenType.DOLLAR_PY_ID)
            opt_value = PyVarExpansion(token)
        elif self._current_token.type_ == lap_lexer.TokenType.PY_STRING:
            token = self._current_token
            self.consume(
                lap_lexer.TokenType.PY_STRING,
                exc_type=lap_exceptions.QueryParamsParseError,
            )
            opt_value = PyScalarValue(token, token.lexeme)
        else:
            opt_value = self.py_dict()

        result = ParamsOption(opt_value)

        return result

    def py_dict(self):
        """Implementation of the ``py_dict`` grammar production rule.

        Production:

            py_dict : LCURL dict_items RCURL
        """
        self.consume(
            lap_lexer.TokenType.LCURL, exc_type=lap_exceptions.QueryParamsParseError
        )
        dict_items = self.dict_items()
        self.consume(
            lap_lexer.TokenType.RCURL, exc_type=lap_exceptions.QueryParamsParseError
        )

        return PyDict(dict_items)

    def dict_items(self):
        """Implementation of the ``dict_items`` grammar production rule.

        Production:

            dict_items : dict_item | (dict_item COMMA dict_items)
        """
        result = []

        item = self.dict_item()
        if item is not None:
            result.append(item)

        while self._current_token.type_ == lap_lexer.TokenType.COMMA:
            self.consume(
                lap_lexer.TokenType.COMMA, exc_type=lap_exceptions.QueryParamsParseError
            )
            item = self.dict_item()
            if item is not None:
                result.append(item)

        return result

    def dict_item(self):
        """Implementation of the ``dict_item`` grammar production rule.

        Production:

            dict_item : (dict_key COLON py_value) | EMPTY
        """
        token = self._current_token

        if token.type_ == lap_lexer.TokenType.PY_STRING:
            key = self.dict_key()
            self.consume(
                lap_lexer.TokenType.COLON, exc_type=lap_exceptions.QueryParamsParseError
            )
            value = self.py_value()
            result = PyDictItem(key, value)
        elif token.type_ == lap_lexer.TokenType.UNKNOWN:
            msg = "Unknown input at position {}: {}".format(token.pos, token.lexeme)
            self.error(msg, exc_type=lap_exceptions.QueryParamsParseError)
        else:
            result = None

        return result

    def dict_key(self):
        """Implementation of the ``dict_key`` grammar production rule.

        Production:

            dict_key : PY_STRING
        """
        token = self._current_token
        self.consume(
            lap_lexer.TokenType.PY_STRING, exc_type=lap_exceptions.QueryParamsParseError
        )
        return PyDictKey(token)

    def py_value(self):
        """Implementation of the ``py_value`` grammar production rule.

        Production:

            py_value : PY_BOOL | PY_NUMBER | PY_STRING | py_tuple | py_list | py_dict
        """
        token = self._current_token

        if token.type_ == lap_lexer.TokenType.PY_BOOL:
            self.consume(
                lap_lexer.TokenType.PY_BOOL,
                exc_type=lap_exceptions.QueryParamsParseError,
            )
            return PyScalarValue(token, token.lexeme)
        elif token.type_ == lap_lexer.TokenType.PY_NUMBER:
            self.consume(
                lap_lexer.TokenType.PY_NUMBER,
                exc_type=lap_exceptions.QueryParamsParseError,
            )
            return PyScalarValue(token, token.lexeme)
        elif token.type_ == lap_lexer.TokenType.PY_STRING:
            self.consume(
                lap_lexer.TokenType.PY_STRING,
                exc_type=lap_exceptions.QueryParamsParseError,
            )
            return PyScalarValue(token, token.lexeme)
        elif token.type_ == lap_lexer.TokenType.LPAREN:
            tuple_node = self.py_tuple()
            return tuple_node
        elif token.type_ == lap_lexer.TokenType.LSQUARE:
            list_node = self.py_list()
            return list_node
        elif token.type_ == lap_lexer.TokenType.LCURL:
            dict_node = self.py_dict()
            return dict_node
        else:
            msg = "Unexpected token type {} at position {}.".format(
                token.type_, token.pos
            )
            self.error(msg, exc_type=lap_exceptions.QueryParamsParseError)

    def py_tuple(self):
        """Implementation of the ``py_tuple`` grammar production rule.

        Production:

            py_tuple : LPAREN collection_items RPAREN
        """
        self.consume(
            lap_lexer.TokenType.LPAREN, exc_type=lap_exceptions.QueryParamsParseError
        )
        items = self.collection_items()
        self.consume(
            lap_lexer.TokenType.RPAREN, exc_type=lap_exceptions.QueryParamsParseError
        )

        return PyTuple(items)

    def py_list(self):
        """Implementation of the ``py_list`` grammar production rule.

        Production:

            py_list : LSQUARE collection_items RSQUARE
        """
        self.consume(
            lap_lexer.TokenType.LSQUARE, exc_type=lap_exceptions.QueryParamsParseError
        )
        items = self.collection_items()
        self.consume(
            lap_lexer.TokenType.RSQUARE, exc_type=lap_exceptions.QueryParamsParseError
        )

        return PyList(items)

    def collection_items(self):
        """Implementation of the ``collection_items`` grammar production rule.

        Production:

            collection_items : collection_item | (collection_item COMMA collection_items)
        """
        result = []

        item = self.collection_item()
        if item is not None:
            result.append(item)

        while self._current_token.type_ == lap_lexer.TokenType.COMMA:
            self.consume(
                lap_lexer.TokenType.COMMA, exc_type=lap_exceptions.QueryParamsParseError
            )
            item = self.collection_item()
            if item is not None:
                result.append(item)

        return result

    def collection_item(self):
        """Implementation of the ``collection_item`` grammar production rule.

        Production:

            collection_item : py_value | EMPTY
        """
        if self._current_token.type_ not in {
            lap_lexer.TokenType.RPAREN,
            lap_lexer.TokenType.RSQUARE,
        }:
            result = self.py_value()
        else:
            result = None  # end of list/tuple items

        return result



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared helper functions for verifying versions of installed modules."""

from typing import Any

from google.cloud.bigquery import exceptions
import packaging.version

_MIN_BQ_STORAGE_VERSION = packaging.version.Version("2.0.0")


class BQStorageVersions:
    """Version comparisons for google-cloud-bigqueyr-storage package."""

    def __init__(self):
        self._installed_version = None

    @property
    def installed_version(self) -> packaging.version.Version:
        """Return the parsed version of google-cloud-bigquery-storage."""
        if self._installed_version is None:
            from google.cloud import bigquery_storage

            self._installed_version = packaging.version.parse(
                # Use 0.0.0, since it is earlier than any released version.
                # Legacy versions also have the same property, but
                # creating a LegacyVersion has been deprecated.
                # https://github.com/pypa/packaging/issues/321
                getattr(bigquery_storage, "__version__", "0.0.0")
            )

        return self._installed_version  # type: ignore

    def try_import(self, raise_if_error: bool = False) -> Any:
        """Tries to import the bigquery_storage module, and returns results
        accordingly. It also verifies the module version is recent enough.

        If the import succeeds, returns the ``bigquery_storage`` module.

        If the import fails,
        returns ``None`` when ``raise_if_error == False``,
        raises Error when ``raise_if_error == True``.

        Returns:
            The ``bigquery_storage`` module or ``None``.

        Raises:
            exceptions.BigQueryStorageNotFoundError:
                If google-cloud-bigquery-storage is not installed
            exceptions.LegacyBigQueryStorageError:
                If google-cloud-bigquery-storage package is outdated
        """
        try:
            from google.cloud import bigquery_storage  # type: ignore
        except ImportError:
            if raise_if_error:
                msg = (
                    "Package google-cloud-bigquery-storage not found. "
                    "Install google-cloud-bigquery-storage version >= "
                    f"{_MIN_BQ_STORAGE_VERSION}."
                )
                raise exceptions.BigQueryStorageNotFoundError(msg)
            return None

        if self.installed_version < _MIN_BQ_STORAGE_VERSION:
            if raise_if_error:
                msg = (
                    "Dependency google-cloud-bigquery-storage is outdated, "
                    f"please upgrade it to version >= {_MIN_BQ_STORAGE_VERSION} "
                    f"(version found: {self.installed_version})."
                )
                raise exceptions.LegacyBigQueryStorageError(msg)
            return None

        return bigquery_storage


BQ_STORAGE_VERSIONS = BQStorageVersions()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for the --pyformat feature."""

from __future__ import annotations

import numbers
import string
from typing import Any


def _field_to_template_value(name: str, value: Any) -> Any:
    """Convert value to something embeddable in a SQL string.

    Does **not** escape strings.
    """
    _validate_type(name, value)
    # TODO(tswast): convert DataFrame objects to gbq tables or literals subquery.
    return value


def _validate_type(name: str, value: Any):
    """Raises TypeError if value is unsupported."""
    if not isinstance(value, (str, numbers.Real)):
        raise TypeError(
            f"{name} has unsupported type: {type(value)}. "
            "Only str, int, float are supported."
        )


def _parse_fields(sql_template: str) -> list[str]:
    return [
        field_name
        for _, field_name, _, _ in string.Formatter().parse(sql_template)
        if field_name is not None
    ]


def pyformat(sql_template: str, user_ns: dict) -> str:
    """Unsafe Python-style string formatting of SQL string.

    Only some data types supported.

    Warning: strings are **not** escaped. This allows them to be used in
    contexts such as table identifiers, where normal query parameters are not
    supported.

    Args:
        sql_template (str):
            SQL string with 0+ {var_name}-style format options.
        user_ns (dict):
            IPython variable namespace to use for formatting.

    Raises:
        TypeError: if a referenced variable is not of a supported type.
        KeyError: if a referenced variable is not found.
    """
    fields = _parse_fields(sql_template)

    format_kwargs = {}
    for name in fields:
        value = user_ns[name]
        format_kwargs[name] = _field_to_template_value(name, value)

    return sql_template.format(**format_kwargs)



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import http.server
import json
import socketserver
import threading
from typing import Dict, List


def execute_node_expansion(params, request):
    return {"error": "Node expansion not yet implemented"}


def convert_graph_data(query_results: Dict[str, Dict[str, str]]):
    """
    Converts graph data to the form expected by the visualization framework.

    Receives graph data as a dictionary, produced by converting the underlying
    DataFrame representing the query results into JSON, then into a
    python dictionary. Converts it into a form expected by the visualization
    framework.

    Args:
        query_results:
            A dictionary with one key/value pair per column. For each column:
            - The key is the name of the column (str)
            - The value is another dictionary with one key/value pair per row.
              Row each row:
              - The key is a string that specifies the integer index of the row
                (e.g. '0', '1', '2')
              - The value is a JSON string containing the result of the query
                for the current row/column. (Note: We only support graph
                visualization for columns of type JSON).
    """
    # Delay spanner imports until this function is called to avoid making
    # spanner-graph-notebook (and its dependencies) hard requirements for bigquery
    # magics users, who don't need graph visualization.
    #
    # Note that these imports do not need to be in a try/except, as this function
    # does not even get called unless spanner_graphs has already been confirmed
    # to exist upstream.
    from spanner_graphs.conversion import get_nodes_edges
    from spanner_graphs.database import SpannerFieldInfo

    try:
        fields: List[SpannerFieldInfo] = []
        data = {}
        tabular_data = {}
        for key, value in query_results.items():
            column_name = None
            column_value = None
            if not isinstance(key, str):
                raise ValueError(f"Expected outer key to be str, got {type(key)}")
            if not isinstance(value, dict):
                raise ValueError(f"Expected outer value to be dict, got {type(value)}")
            column_name = key
            column_value = value

            fields.append(SpannerFieldInfo(name=column_name, typename="JSON"))
            data[column_name] = []
            tabular_data[column_name] = []
            for value_key, value_value in column_value.items():
                try:
                    row_json = json.loads(value_value)
                    data[column_name].append(row_json)
                    tabular_data[column_name].append(row_json)
                except (ValueError, TypeError):
                    # Non-JSON columns cannot be visualized, but we still want them
                    # in the tabular view.
                    tabular_data[column_name].append(str(value_value))

        nodes, edges = get_nodes_edges(data, fields, schema_json=None)

        # Convert nodes and edges to json objects.
        # (Unfortunately, the code coverage tooling does not allow this
        #  to be expressed as list comprehension).
        nodes_json = []
        for node in nodes:
            nodes_json.append(node.to_json())
        edges_json = []
        for edge in edges:
            edges_json.append(edge.to_json())

        return {
            "response": {
                # These fields populate the graph result view.
                "nodes": nodes_json,
                "edges": edges_json,
                # This populates the visualizer's schema view, but not yet implemented on the
                # BigQuery side.
                "schema": None,
                # This field is used to populate the visualizer's tabular view.
                "query_result": tabular_data,
            }
        }
    except Exception as e:
        return {"error": getattr(e, "message", str(e))}


class GraphServer:
    """
    Http server invoked by Javascript to obtain the query results for visualization.

    The server is invoked by Javascript, generated as part of
    spanner_graphs.graph_visualization.generate_visualization_html().

    This server is used only in Jupyter; in colab, google.colab.output.register_callback()
    is used instead.
    """

    host = "http://localhost"
    endpoints = {
        "get_ping": "/get_ping",
        "post_ping": "/post_ping",
        "post_node_expansion": "/post_node_expansion",
        "post_query": "/post_query",
    }

    def __init__(self):
        self.port = None
        self.url = None
        self._server = None

    def build_route(self, endpoint):
        """
        Returns a url for connecting to the given endpoint.
        Supported values include:
          - "get_ping": sends a GET request to ping the server.
          - "post_ping": sends a POST request to ping the server.
          - "post_query": sends a POST request to obtain query results.
        """
        return f"{self.url}{endpoint}"

    def _start_server(self):
        class ThreadedTCPServer(socketserver.TCPServer):
            # Allow socket reuse to avoid "Address already in use" errors
            allow_reuse_address = True
            # Daemon threads automatically terminate when the main program exits
            daemon_threads = True

        with ThreadedTCPServer(("", self.port), GraphServerHandler) as httpd:
            self._server = httpd
            self._server.serve_forever()

    def init(self):
        """
        Starts the HTTP server. The server runs forever, until stop_server() is called.
        """
        import portpicker

        self.port = portpicker.pick_unused_port()
        self.url = f"{GraphServer.host}:{self.port}"

        server_thread = threading.Thread(target=self._start_server)
        server_thread.start()
        return server_thread

    def stop_server(self):
        """
        Starts the HTTP server, if it is currently running.
        """
        if self._server:
            self._server.shutdown()
            print("BigQuery-magics graph server shutting down...")
            self._server = None


global graph_server
graph_server = GraphServer()


class GraphServerHandler(http.server.SimpleHTTPRequestHandler):
    """
    Handles HTTP requests send to the graph server.
    """

    def log_message(self, format, *args):
        pass

    def do_json_response(self, data):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Content-type", "application/json")
        self.send_header("Access-Control-Allow-Methods", "GET,PUT,POST,DELETE,OPTIONS")
        self.end_headers()
        self.wfile.write(json.dumps(data).encode())

    def do_message_response(self, message):
        self.do_json_response({"message": message})

    def do_data_response(self, data):
        self.do_json_response(data)

    def parse_post_data(self):
        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length).decode("utf-8")
        return json.loads(post_data)

    def handle_get_ping(self):
        self.do_message_response("pong")

    def handle_post_ping(self):
        data = self.parse_post_data()
        self.do_data_response({"your_request": data})

    def handle_post_query(self):
        data = self.parse_post_data()
        response = convert_graph_data(query_results=json.loads(data["params"]))
        self.do_data_response(response)

    def handle_post_node_expansion(self):
        """Handle POST requests for node expansion.

        Expects a JSON payload with:
        - params: A JSON string containing connection parameters (project, instance, database, graph)
        - request: A dictionary with node details (uid, node_labels, node_properties, direction, edge_label)
        """
        data = self.parse_post_data()

        # Execute node expansion with:
        # - params_str: JSON string with connection parameters (project, instance, database, graph)
        # - request: Dict with node details (uid, node_labels, node_properties, direction, edge_label)
        self.do_data_response(
            execute_node_expansion(
                params=data.get("params"), request=data.get("request")
            )
        )

    def do_GET(self):
        assert self.path == GraphServer.endpoints["get_ping"]
        self.handle_get_ping()

    def do_POST(self):
        if self.path == GraphServer.endpoints["post_ping"]:
            self.handle_post_ping()
        elif self.path == GraphServer.endpoints["post_node_expansion"]:
            self.handle_post_node_expansion()
        else:
            assert self.path == GraphServer.endpoints["post_query"]
            self.handle_post_query()


atexit.register(graph_server.stop_server)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bigquery_magics.config
import bigquery_magics.version

context = bigquery_magics.config.context
__version__ = bigquery_magics.version.__version__

# Whether the magics has already been reigstered by some other packages.
is_registered = False


def load_ipython_extension(ipython):
    """Called by IPython when this module is loaded as an IPython extension."""
    # Import here to avoid circular imports.
    from bigquery_magics.bigquery import _cell_magic

    ipython.register_magic_function(
        _cell_magic, magic_kind="cell", magic_name="bigquery"
    )
    ipython.register_magic_function(_cell_magic, magic_kind="cell", magic_name="bqsql")

    global is_registered
    is_registered = True


def unload_ipython_extension(ipython):
    global is_registered
    is_registered = False


__all__ = (
    # For backwards compatibility we need to make the context available in
    # the path google.cloud.bigquery.magics.context.
    "context",
    "__version__",
    "load_ipython_extension",
    "unload_ipython_extension",
    "is_registered",
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest.mock as mock

import google.auth.credentials
import pydata_google_auth
import pytest

import bigquery_magics


def test_context_with_default_credentials():
    """When Application Default Credentials are set, the context credentials
    will be created the first time it is called
    """
    bigquery_magics.context._credentials = None
    bigquery_magics.context._project = None

    project = "prahj-ekt"
    credentials_mock = mock.create_autospec(
        google.auth.credentials.Credentials, instance=True
    )
    default_patch = mock.patch.object(
        pydata_google_auth,
        "default",
        autospec=True,
        return_value=(credentials_mock, project),
    )
    with default_patch as default_mock:
        assert bigquery_magics.context.credentials is credentials_mock
        assert bigquery_magics.context.project == project

    assert default_mock.call_count == 2


def test_context_credentials_and_project_can_be_set_explicitly():
    project1 = "one-project-55564"
    project2 = "other-project-52569"
    credentials_mock = mock.create_autospec(
        google.auth.credentials.Credentials, instance=True
    )
    default_patch = mock.patch.object(
        pydata_google_auth,
        "default",
        autospec=True,
        return_value=(credentials_mock, project1),
    )
    with default_patch as default_mock:
        bigquery_magics.context.credentials = credentials_mock
        bigquery_magics.context.project = project2

    assert bigquery_magics.context.project == project2
    assert bigquery_magics.context.credentials is credentials_mock
    # default should not be called if credentials & project are explicitly set
    assert default_mock.call_count == 0


def test_context_set_default_variable():
    assert bigquery_magics.context.default_variable is None

    bigquery_magics.context.default_variable = "_bq_df"
    assert bigquery_magics.context.default_variable == "_bq_df"

    bigquery_magics.context.default_variable = None
    assert bigquery_magics.context.default_variable is None


@pytest.mark.parametrize("engine", ["pandas", "bigframes"])
def test_context_set_engine(monkeypatch, engine):
    monkeypatch.setattr(bigquery_magics.context, "engine", engine)
    assert bigquery_magics.context.engine == engine


def test_context_set_invalid_engine(monkeypatch):
    with pytest.raises(ValueError):
        monkeypatch.setattr(bigquery_magics.context, "engine", "whatever")



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
import json
import os
import pathlib

Path = pathlib.Path


# The identifier for GCP VS Code extension
# https://cloud.google.com/code/docs/vscode/install
GOOGLE_CLOUD_CODE_EXTENSION_NAME = "googlecloudtools.cloudcode"


# The identifier for BigQuery Jupyter notebook plugin
# https://cloud.google.com/bigquery/docs/jupyterlab-plugin
BIGQUERY_JUPYTER_PLUGIN_NAME = "bigquery_jupyter_plugin"


def _is_vscode_extension_installed(extension_id: str) -> bool:
    """
    Checks if a given Visual Studio Code extension is installed.

    Args:
        extension_id: The ID of the extension (e.g., "ms-python.python").

    Returns:
        True if the extension is installed, False otherwise.
    """
    try:
        # Determine the user's VS Code extensions directory.
        user_home = Path.home()
        vscode_extensions_dir = user_home / ".vscode" / "extensions"

        # Check if the extensions directory exists.
        if not vscode_extensions_dir.exists():
            return False

        # Iterate through the subdirectories in the extensions directory.
        extension_dirs = filter(
            lambda p: p.is_dir() and p.name.startswith(extension_id + "-"),
            vscode_extensions_dir.iterdir(),
        )
        for extension_dir in extension_dirs:
            # As a more robust check, the manifest file must exist.
            manifest_path = extension_dir / "package.json"
            if not manifest_path.exists() or not manifest_path.is_file():
                continue

            # Finally, the manifest file must be a valid json
            with open(manifest_path, "r", encoding="utf-8") as f:
                json.load(f)

            return True
    except Exception:
        pass

    return False


def _is_package_installed(package_name: str) -> bool:
    """
    Checks if a Python package is installed.

    Args:
        package_name: The name of the package to check (e.g., "requests", "numpy").

    Returns:
        True if the package is installed, False otherwise.
    """
    try:
        importlib.import_module(package_name)
        return True
    except Exception:
        return False


def is_vscode() -> bool:
    return os.getenv("VSCODE_PID") is not None


def is_jupyter() -> bool:
    return os.getenv("JPY_PARENT_PID") is not None


def is_vscode_google_cloud_code_extension_installed() -> bool:
    return _is_vscode_extension_installed(GOOGLE_CLOUD_CODE_EXTENSION_NAME)


def is_jupyter_bigquery_plugin_installed() -> bool:
    return _is_package_installed(BIGQUERY_JUPYTER_PLUGIN_NAME)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bigquery_magics.line_arg_parser.exceptions import (
    DuplicateQueryParamsError,
    ParseError,
    QueryParamsParseError,
)
from bigquery_magics.line_arg_parser.lexer import Lexer, TokenType
from bigquery_magics.line_arg_parser.parser import Parser
from bigquery_magics.line_arg_parser.visitors import QueryParamsExtractor

__all__ = (
    "DuplicateQueryParamsError",
    "Lexer",
    "Parser",
    "ParseError",
    "QueryParamsExtractor",
    "QueryParamsParseError",
    "TokenType",
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

IPython = pytest.importorskip("IPython")


@pytest.fixture
def base_visitor():
    from bigquery_magics.line_arg_parser.visitors import NodeVisitor

    return NodeVisitor()


def test_unknown_node(base_visitor):
    from bigquery_magics.line_arg_parser.parser import ParseNode

    class UnknownNode(ParseNode):
        pass

    node = UnknownNode()

    with pytest.raises(Exception, match=r"No visit_UnknownNode method"):
        base_visitor.visit(node)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import pytest

try:
    from google.cloud import bigquery_storage  # type: ignore
except ImportError:
    bigquery_storage = None

from google.cloud.bigquery import exceptions

from bigquery_magics import _versions_helpers


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_raises_no_error_w_recent_bqstorage():
    with mock.patch("google.cloud.bigquery_storage.__version__", new="2.0.0"):
        try:
            bqstorage_versions = _versions_helpers.BQStorageVersions()
            bqstorage_versions.try_import(raise_if_error=True)
        except exceptions.LegacyBigQueryStorageError:  # pragma: NO COVER
            raise ("Legacy error raised with a non-legacy dependency version.")


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_raises_error_w_legacy_bqstorage():
    with mock.patch("google.cloud.bigquery_storage.__version__", new="1.9.9"):
        with pytest.raises(exceptions.LegacyBigQueryStorageError):
            bqstorage_versions = _versions_helpers.BQStorageVersions()
            bqstorage_versions.try_import(raise_if_error=True)


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_returns_none_with_legacy_bqstorage():
    with mock.patch("google.cloud.bigquery_storage.__version__", new="1.9.9"):
        try:
            bqstorage_versions = _versions_helpers.BQStorageVersions()
            bq_storage = bqstorage_versions.try_import()
        except exceptions.LegacyBigQueryStorageError:  # pragma: NO COVER
            raise ("Legacy error raised when raise_if_error == False.")
        assert bq_storage is None


@pytest.mark.skipif(
    bigquery_storage is not None,
    reason="Tests behavior when `google-cloud-bigquery-storage` isn't installed",
)
def test_returns_none_with_bqstorage_uninstalled():
    try:
        bqstorage_versions = _versions_helpers.BQStorageVersions()
        bq_storage = bqstorage_versions.try_import()
    except exceptions.LegacyBigQueryStorageError:  # pragma: NO COVER
        raise ("NotFound error raised when raise_if_error == False.")
    assert bq_storage is None


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_raises_error_w_unknown_bqstorage_version():
    with mock.patch("google.cloud.bigquery_storage", autospec=True) as fake_module:
        del fake_module.__version__
        error_pattern = r"version found: 0.0.0"
        with pytest.raises(exceptions.LegacyBigQueryStorageError, match=error_pattern):
            bqstorage_versions = _versions_helpers.BQStorageVersions()
            bqstorage_versions.try_import(raise_if_error=True)


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_installed_bqstorage_version_returns_cached():
    bqstorage_versions = _versions_helpers.BQStorageVersions()
    bqstorage_versions._installed_version = object()
    assert bqstorage_versions.installed_version is bqstorage_versions._installed_version


@pytest.mark.skipif(
    bigquery_storage is None, reason="Requires `google-cloud-bigquery-storage`"
)
def test_installed_bqstorage_version_returns_parsed_version():
    bqstorage_versions = _versions_helpers.BQStorageVersions()
    with mock.patch("google.cloud.bigquery_storage.__version__", new="1.2.3"):
        bqstorage_versions = bqstorage_versions.installed_version

    assert bqstorage_versions.major == 1
    assert bqstorage_versions.minor == 2
    assert bqstorage_versions.micro == 3



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional

import google.api_core.client_options as client_options
import google.cloud.bigquery as bigquery
import pydata_google_auth

_SCOPES = ["https://www.googleapis.com/auth/cloud-platform"]


def _get_default_credentials_with_project():
    return pydata_google_auth.default(scopes=_SCOPES, use_local_webserver=False)


@dataclass
class Context(object):
    """Storage for objects to be used throughout an IPython notebook session.

    A Context object is initialized when the ``bigquery_magics`` module is imported,
    and can be found at ``bigquery_magics.context``.
    """

    _connection = None

    default_query_job_config = bigquery.QueryJobConfig()
    """google.cloud.bigquery.job.QueryJobConfig: Default job
        configuration for queries.

        The context's :class:`~google.cloud.bigquery.job.QueryJobConfig` is
        used for queries. Some properties can be overridden with arguments to
        the magics.

        Example:
            Manually setting the default value for ``maximum_bytes_billed``
            to 100 MB:

            >>> from google.cloud.bigquery import magics
            >>> bigquery_magics.context.default_query_job_config.maximum_bytes_billed = 100000000
    """

    bigquery_client_options = client_options.ClientOptions()
    """google.api_core.client_options.ClientOptions: client options to be
        used through IPython magics.

        Note::
            The client options do not need to be explicitly defined if no
            special network connections are required. Normally you would be
            using the https://bigquery.googleapis.com/ end point.

        Example:
            Manually setting the endpoint:

            >>> from google.cloud.bigquery import magics
            >>> client_options = {}
            >>> client_options['api_endpoint'] = "https://some.special.url"
            >>> bigquery_magics.context.bigquery_client_options = client_options
    """

    bqstorage_client_options = client_options.ClientOptions()
    """google.api_core.client_options.ClientOptions: client options to be
        used through IPython magics for the storage client.

        Note::
            The client options do not need to be explicitly defined if no
            special network connections are required. Normally you would be
            using the https://bigquerystorage.googleapis.com/ end point.

        Example:
            Manually setting the endpoint:

            >>> from google.cloud.bigquery import magics
            >>> client_options = {}
            >>> client_options['api_endpoint'] = "https://some.special.url"
            >>> bigquery_magics.context.bqstorage_client_options = client_options
    """

    progress_bar_type = "tqdm_notebook"
    """str: Default progress bar type to use to display progress bar while
        executing queries through IPython magics.

        Note::
            Install the ``tqdm`` package to use this feature.

        Example:
            Manually setting the progress_bar_type:

            >>> from google.cloud.bigquery import magics
            >>> bigquery_magics.context.progress_bar_type = "tqdm_notebook"
    """

    _credentials = None

    @property
    def credentials(self):
        """google.auth.credentials.Credentials: Credentials to use for queries
        performed through IPython magics.

        Note:
            These credentials do not need to be explicitly defined if you are
            using Application Default Credentials. If you are not using
            Application Default Credentials, manually construct a
            :class:`google.auth.credentials.Credentials` object and set it as
            the context credentials as demonstrated in the example below. See
            `auth docs`_ for more information on obtaining credentials.

        Example:
            Manually setting the context credentials:

            >>> from google.cloud.bigquery import magics
            >>> from google.oauth2 import service_account
            >>> credentials = (service_account
            ...     .Credentials.from_service_account_file(
            ...         '/path/to/key.json'))
            >>> bigquery_magics.context.credentials = credentials


        .. _auth docs: http://google-auth.readthedocs.io
            /en/latest/user-guide.html#obtaining-credentials
        """
        if self._credentials is None:
            self._credentials, _ = _get_default_credentials_with_project()
        return self._credentials

    @credentials.setter
    def credentials(self, value):
        self._credentials = value

    _default_variable: Optional[str] = None

    @property
    def default_variable(self) -> Optional[str]:
        """Optional[str]: If set, the variable name to use for the DataFrame
        returned from running the query.
        """
        return self._default_variable

    @default_variable.setter
    def default_variable(self, value: Optional[str]):
        self._default_variable = value

    _project = None

    @property
    def project(self):
        """str: Default project to use for queries performed through IPython
        magics.

        Note:
            The project does not need to be explicitly defined if you have an
            environment default project set. If you do not have a default
            project set in your environment, manually assign the project as
            demonstrated in the example below.

        Example:
            Manually setting the context project:

            >>> from google.cloud.bigquery import magics
            >>> bigquery_magics.context.project = 'my-project'
        """
        if self._project is None:
            _, self._project = _get_default_credentials_with_project()
        return self._project

    @project.setter
    def project(self, value):
        self._project = value

    _engine = "pandas"

    @property
    def engine(self) -> str:
        """Engine to run the query. Could either be "pandas" or "bigframes".

        If using "pandas", the query result will be stored in a Pandas dataframe.
        If using "bigframes", the query result will be stored in a bigframes dataframe instead.

        Example:
            Manully setting the content engine:

            >>> from google.cloud.bigquery import magics
            >>> bigquery_magics.context.engine = 'bigframes'
        """
        return self._engine

    @engine.setter
    def engine(self, value):
        if value != "pandas" and value != "bigframes":
            raise ValueError("engine must be either 'pandas' or 'bigframes'")
        self._engine = value


context = Context()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""System tests for Jupyter/IPython connector."""

import re

from IPython.testing import globalipapp
from IPython.utils import io
import pandas
import psutil


def test_bigquery_magic():
    globalipapp.start_ipython()
    ip = globalipapp.get_ipython()
    current_process = psutil.Process()
    conn_count_start = len(current_process.net_connections())

    ip.extension_manager.load_extension("bigquery_magics")
    sql = """
        SELECT
            CONCAT(
            'https://stackoverflow.com/questions/',
            CAST(id as STRING)) as url,
            view_count
        FROM `bigquery-public-data.stackoverflow.posts_questions`
        WHERE tags like '%google-bigquery%'
        ORDER BY view_count DESC
        LIMIT 10
    """
    with io.capture_output() as captured:
        result = ip.run_cell_magic("bigquery", "--use_rest_api", sql)

    conn_count_end = len(current_process.net_connections())

    lines = re.split("\n|\r", captured.stdout)
    # Removes blanks & terminal code (result of display clearing)
    updates = list(filter(lambda x: bool(x) and x != "\x1b[2K", lines))
    assert re.match("Executing query with job ID: .*", updates[0])
    assert (re.match("Query executing: .*s", line) for line in updates[1:-1])
    assert isinstance(result, pandas.DataFrame)
    assert len(result) == 10  # verify row count
    assert list(result) == ["url", "view_count"]  # verify column names

    # NOTE: For some reason, the number of open sockets is sometimes one *less*
    # than expected when running system tests on Kokoro, thus using the <= assertion.
    # That's still fine, however, since the sockets are apparently not leaked.
    assert conn_count_end <= conn_count_start  # system resources are released



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

IPython = pytest.importorskip("IPython")


@pytest.fixture(scope="session")
def lexer_class():
    from bigquery_magics.line_arg_parser.lexer import Lexer

    return Lexer


def test_empy_input(lexer_class):
    from bigquery_magics.line_arg_parser import TokenType
    from bigquery_magics.line_arg_parser.lexer import Token

    lexer = lexer_class("")
    tokens = list(lexer)

    assert tokens == [Token(TokenType.EOL, lexeme="", pos=0)]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import google.auth.credentials
import pytest
import test_utils.imports  # google-cloud-testutils

import bigquery_magics


@pytest.fixture()
def ipython_ns_cleanup():
    """A helper to clean up user namespace after the test

    for the duration of the test scope.
    """
    names_to_clean = []  # pairs (IPython_instance, name_to_clean)

    yield names_to_clean

    for ip, name in names_to_clean:
        if name in ip.user_ns:
            del ip.user_ns[name]


@pytest.fixture(scope="session")
def missing_bq_storage():
    """Provide a patcher that can make the bigquery storage import to fail."""

    def fail_if(name, globals, locals, fromlist, level):
        # NOTE: *very* simplified, assuming a straightforward absolute import
        return "bigquery_storage" in name or (
            fromlist is not None and "bigquery_storage" in fromlist
        )

    return test_utils.imports.maybe_fail_import(predicate=fail_if)


@pytest.fixture(scope="session")
def missing_grpcio_lib():
    """Provide a patcher that can make the gapic library import to fail."""

    def fail_if(name, globals, locals, fromlist, level):
        # NOTE: *very* simplified, assuming a straightforward absolute import
        return "gapic_v1" in name or (fromlist is not None and "gapic_v1" in fromlist)

    return test_utils.imports.maybe_fail_import(predicate=fail_if)


@pytest.fixture
def mock_credentials(monkeypatch):
    credentials = mock.create_autospec(
        google.auth.credentials.Credentials, instance=True
    )

    # Set up the context with monkeypatch so that it's reset for subsequent
    # tests.
    monkeypatch.setattr(bigquery_magics.context, "_project", "test-project")
    monkeypatch.setattr(bigquery_magics.context, "_credentials", credentials)


@pytest.fixture
def set_bigframes_engine_in_context(monkeypatch):
    monkeypatch.setattr(bigquery_magics.context, "engine", "bigframes")



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

import setuptools

# Package metadata.

name = "bigquery-magics"
description = "Google BigQuery magics for Jupyter and IPython"

# Should be one of:
# 'Development Status :: 3 - Alpha'
# 'Development Status :: 4 - Beta'
# 'Development Status :: 5 - Production/Stable'``
release_status = "Development Status :: 4 - Beta"
dependencies = [
    "db-dtypes>=0.3.0,<2.0.0",
    "google-cloud-bigquery >= 3.13.0, <4.0.0",
    "ipywidgets>=7.7.1",
    "ipython>=7.23.1",
    "ipykernel>=5.5.6",
    "packaging >= 20.0.0",
    "pandas>=1.2.0",
    "pyarrow >= 3.0.0",
    "pydata-google-auth >=1.5.0",
    "tqdm >= 4.7.4, <5.0.0",
]
extras = {
    # bqstorage had a period where it was a required dependency, and has been
    # moved back to optional due to bloat.  See
    # https://github.com/googleapis/python-bigquery/issues/1196 for more background.
    "bqstorage": [
        "google-cloud-bigquery-storage >= 2.6.0, <3.0.0",
        # Due to an issue in pip's dependency resolver, the `grpc` extra is not
        # installed, even though `google-cloud-bigquery-storage` specifies it
        # as `google-api-core[grpc]`. We thus need to explicitly specify it here.
        # See: https://github.com/googleapis/python-bigquery/issues/83 The
        # grpc.Channel.close() method isn't added until 1.32.0.
        # https://github.com/grpc/grpc/pull/15254
        "grpcio >= 1.47.0, < 2.0.0",
        "grpcio >= 1.49.1, < 2.0.0; python_version>='3.11'",
    ],
    "bigframes": ["bigframes >= 1.17.0"],
    "geopandas": ["geopandas >= 1.0.1"],
    "spanner-graph-notebook": [
        "spanner-graph-notebook >= 1.1.7",
        "portpicker",
    ],
}

all_extras = []

for extra in extras:
    all_extras.extend(extras[extra])

extras["all"] = all_extras

# Setup boilerplate below this line.

package_root = os.path.abspath(os.path.dirname(__file__))

readme_filename = os.path.join(package_root, "README.rst")
with io.open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()

version = {}
with open(os.path.join(package_root, "bigquery_magics/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

# Only include packages under the 'bigquery_magics' namespace. Do not include tests,
# benchmarks, etc.
packages = [
    package
    for package in setuptools.find_namespace_packages()
    if package.startswith("bigquery_magics")
]

setuptools.setup(
    name=name,
    version=version,
    description=description,
    long_description=readme,
    author="Google LLC",
    author_email="googleapis-packages@google.com",
    license="Apache 2.0",
    url="https://github.com/googleapis/python-bigquery-magics",
    classifiers=[
        release_status,
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Operating System :: OS Independent",
        "Topic :: Internet",
    ],
    platforms="Posix; MacOS X; Windows",
    packages=packages,
    install_requires=dependencies,
    extras_require=extras,
    python_requires=">=3.9",
    include_package_data=True,
    zip_safe=False,
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IPython Magics

.. function:: ``%%bigquery`` or ``%%bqsql``

    IPython cell magic to run a query and display the result as a DataFrame

    .. code-block:: python

        %%bqsql [<destination_var>] [--project <project>] [--use_legacy_sql]
                   [--verbose] [--params <params>]
        <query>

    Parameters:

    * ``<destination_var>`` (Optional[line argument]):
        variable to store the query results. The results are not displayed if
        this parameter is used. If an error occurs during the query execution,
        the corresponding ``QueryJob`` instance (if available) is stored in
        the variable instead. Set ``bigquery_magics.context.default_variable``
        to set a destination variable without specifying this argument.
    * ``--destination_table`` (Optional[line argument]):
        A dataset and table to store the query results. If table does not exists,
        it will be created. If table already exists, its data will be overwritten.
        Variable should be in a format <dataset_id>.<table_id>.
    * ``--no_query_cache`` (Optional[line argument]):
        Do not use cached query results.
    * ``--project <project>`` (Optional[line argument]):
        Project to use for running the query. Defaults to the context
        :attr:`~google.cloud.bigquery.magics.Context.project`.
    * ``--use_bqstorage_api`` (Optional[line argument]):
        [Deprecated] Not used anymore, as BigQuery Storage API is used by default.
    * ``--use_rest_api`` (Optional[line argument]):
        Use the BigQuery REST API instead of the Storage API.
    * ``--use_legacy_sql`` (Optional[line argument]):
        Runs the query using Legacy SQL syntax. Defaults to Standard SQL if
        this argument not used.
    * ``--verbose`` (Optional[line argument]):
        If this flag is used, information including the query job ID and the
        amount of time for the query to complete will not be cleared after the
        query is finished. By default, this information will be displayed but
        will be cleared after the query is finished.
    * ``--graph`` (Optional[line argument]):
        Visualizes the query result as a graph.
    * ``--use_geodataframe <params>`` (Optional[line argument]):
        Return the query result as a geopandas.GeoDataFrame.
        If present, the argument that follows the ``--use_geodataframe`` flag
        must be a string representing column names to use as the active
        geometry.

        See geopandas.GeoDataFrame for details.
        The Coordinate Reference System will be set to “EPSG:4326”.
    * ``--params <params>`` (Optional[line argument]):
        If present, the argument following the ``--params`` flag must be
        either:

        * :class:`str` - A JSON string representation of a dictionary in the
          format ``{"param_name": "param_value"}`` (ex. ``{"num": 17}``). Use
          of the parameter in the query should be indicated with
          ``@param_name``. See ``In[5]`` in the Examples section below.

        * :class:`dict` reference - A reference to a ``dict`` in the format
          ``{"param_name": "param_value"}``, where the value types must be JSON
          serializable. The variable reference is indicated by a ``$`` before
          the variable name (ex. ``$my_dict_var``). See ``In[6]`` and ``In[7]``
          in the Examples section below.
    * ``--engine <engine>`` (Optional[line argument]):
          Set the execution engine, either 'pandas' (default) or 'bigframes'
          (experimental).
    * ``--pyformat`` (Optional[line argument]):
          Warning! Do not use with user-provided values.
          This doesn't escape values. Use --params instead for proper SQL escaping.
          This enables Python string formatting in the query text.
          Useful for values not supported by SQL query params such as table IDs.

    * ``<query>`` (required, cell argument):
        SQL query to run. If the query does not contain any whitespace (aside
        from leading and trailing whitespace), it is assumed to represent a
        fully-qualified table ID, and the latter's data will be fetched.

    Returns:
        A :class:`pandas.DataFrame` or :class:`bigframes.pandas.DataFrame`
        with the query results, depending on the ``engine`` chosen or if
        ``--as_geodataframe`` was provided.

    .. note::
        All queries run using this magic will run using the context
        :attr:`~bigquery_magics.config.Context.credentials`.
"""

from __future__ import print_function

import ast
from concurrent import futures
import copy
import json
import re
import sys
import threading
import time
from typing import Any, List, Tuple
import warnings

import IPython  # type: ignore
from IPython.core import magic_arguments  # type: ignore
from IPython.core.getipython import get_ipython
from google.api_core import client_info
from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import exceptions
from google.cloud.bigquery.dataset import DatasetReference
from google.cloud.bigquery.dbapi import _helpers
from google.cloud.bigquery.job import QueryJobConfig
import pandas

from bigquery_magics import environment
from bigquery_magics import line_arg_parser as lap
import bigquery_magics._versions_helpers
import bigquery_magics.config
import bigquery_magics.graph_server as graph_server
import bigquery_magics.pyformat
import bigquery_magics.version

try:
    from google.cloud import bigquery_storage  # type: ignore
except ImportError:
    bigquery_storage = None

try:
    import bigframes.pandas as bpd
except ImportError:
    bpd = None

context = bigquery_magics.config.context


def _get_user_agent():
    identities = [
        f"ipython-{IPython.__version__}",
        f"bigquery-magics/{bigquery_magics.version.__version__}",
    ]

    if environment.is_vscode():
        identities.append("vscode")
        if environment.is_vscode_google_cloud_code_extension_installed():
            identities.append(environment.GOOGLE_CLOUD_CODE_EXTENSION_NAME)
    elif environment.is_jupyter():
        identities.append("jupyter")
        if environment.is_jupyter_bigquery_plugin_installed():
            identities.append(environment.BIGQUERY_JUPYTER_PLUGIN_NAME)

    return " ".join(identities)


def _handle_error(error, destination_var=None):
    """Process a query execution error.

    Args:
        error (Exception):
            An exception that occurred during the query execution.
        destination_var (Optional[str]):
            The name of the IPython session variable to store the query job.
    """
    if destination_var:
        query_job = getattr(error, "query_job", None)

        if query_job is not None:
            get_ipython().push({destination_var: query_job})
        else:
            # this is the case when previewing table rows by providing just
            # table ID to cell magic
            print(
                "Could not save output to variable '{}'.".format(destination_var),
                file=sys.stderr,
            )

    print("\nERROR:\n", str(error), file=sys.stderr)


def _run_query(client, query, job_config=None):
    """Runs a query while printing status updates

    Args:
        client (google.cloud.bigquery.client.Client):
            Client to bundle configuration needed for API requests.
        query (str):
            SQL query to be executed. Defaults to the standard SQL dialect.
            Use the ``job_config`` parameter to change dialects.
        job_config (Optional[google.cloud.bigquery.job.QueryJobConfig]):
            Extra configuration options for the job.

    Returns:
        google.cloud.bigquery.job.QueryJob: the query job created

    Example:
        >>> client = bigquery.Client()
        >>> _run_query(client, "SELECT 17")
        Executing query with job ID: bf633912-af2c-4780-b568-5d868058632b
        Query executing: 1.66s
        Query complete after 2.07s
        'bf633912-af2c-4780-b568-5d868058632b'
    """
    start_time = time.perf_counter()
    query_job = client.query(query, job_config=job_config)

    if job_config and job_config.dry_run:
        return query_job

    print(f"Executing query with job ID: {query_job.job_id}")

    while True:
        print(
            f"\rQuery executing: {time.perf_counter() - start_time:.2f}s".format(),
            end="",
        )
        try:
            query_job.result(timeout=0.5)
            break
        except futures.TimeoutError:
            continue
    print(f"\nJob ID {query_job.job_id} successfully executed")
    return query_job


def _create_dataset_if_necessary(client, dataset_id):
    """Create a dataset in the current project if it doesn't exist.

    Args:
        client (google.cloud.bigquery.client.Client):
            Client to bundle configuration needed for API requests.
        dataset_id (str):
            Dataset id.
    """
    dataset_reference = DatasetReference(client.project, dataset_id)
    try:
        dataset = client.get_dataset(dataset_reference)
        return
    except NotFound:
        pass
    dataset = bigquery.Dataset(dataset_reference)
    dataset.location = client.location
    print(f"Creating dataset: {dataset_id}")
    dataset = client.create_dataset(dataset)


@magic_arguments.magic_arguments()
@magic_arguments.argument(
    "destination_var",
    nargs="?",
    help=("If provided, save the output to this variable instead of displaying it."),
)
@magic_arguments.argument(
    "--destination_table",
    type=str,
    default=None,
    help=(
        "If provided, save the output of the query to a new BigQuery table. "
        "Variable should be in a format <dataset_id>.<table_id>. "
        "If table does not exists, it will be created. "
        "If table already exists, its data will be overwritten."
    ),
)
@magic_arguments.argument(
    "--project",
    type=str,
    default=None,
    help=("Project to use for executing this query. Defaults to the context project."),
)
@magic_arguments.argument(
    "--max_results",
    default=None,
    help=(
        "Maximum number of rows in dataframe returned from executing the query."
        "Defaults to returning all rows."
    ),
)
@magic_arguments.argument(
    "--maximum_bytes_billed",
    default=None,
    help=(
        "maximum_bytes_billed to use for executing this query. Defaults to "
        "the context default_query_job_config.maximum_bytes_billed."
    ),
)
@magic_arguments.argument(
    "--dry_run",
    action="store_true",
    default=False,
    help=(
        "Sets query to be a dry run to estimate costs. "
        "Defaults to executing the query instead of dry run if this argument is not used."
        "Does not work with engine 'bigframes'. "
    ),
)
@magic_arguments.argument(
    "--use_legacy_sql",
    action="store_true",
    default=False,
    help=(
        "Sets query to use Legacy SQL instead of Standard SQL. Defaults to "
        "Standard SQL if this argument is not used."
    ),
)
@magic_arguments.argument(
    "--bigquery_api_endpoint",
    type=str,
    default=None,
    help=(
        "The desired API endpoint, e.g., bigquery.googlepis.com. Defaults to this "
        "option's value in the context bigquery_client_options."
    ),
)
@magic_arguments.argument(
    "--bqstorage_api_endpoint",
    type=str,
    default=None,
    help=(
        "The desired API endpoint, e.g., bigquerystorage.googlepis.com. Defaults to "
        "this option's value in the context bqstorage_client_options."
    ),
)
@magic_arguments.argument(
    "--no_query_cache",
    action="store_true",
    default=False,
    help=("Do not use cached query results."),
)
@magic_arguments.argument(
    "--use_bqstorage_api",
    action="store_true",
    default=None,
    help=(
        "[Deprecated] The BigQuery Storage API is already used by default to "
        "download large query results, and this option has no effect. "
        "If you want to switch to the classic REST API instead, use the "
        "--use_rest_api option."
    ),
)
@magic_arguments.argument(
    "--use_rest_api",
    action="store_true",
    default=False,
    help=(
        "Use the classic REST API instead of the BigQuery Storage API to "
        "download query results."
    ),
)
@magic_arguments.argument(
    "--verbose",
    action="store_true",
    default=False,
    help=(
        "If set, print verbose output, including the query job ID and the "
        "amount of time for the query to finish. By default, this "
        "information will be displayed as the query runs, but will be "
        "cleared after the query is finished."
        "This flag is ignored when the engine is 'bigframes'."
    ),
)
@magic_arguments.argument(
    "--params",
    nargs="+",
    default=None,
    help=(
        "Parameters to format the query string. If present, the --params "
        "flag should be followed by a string representation of a dictionary "
        "in the format {'param_name': 'param_value'} (ex. {\"num\": 17}), "
        "or a reference to a dictionary in the same format. The dictionary "
        "reference can be made by including a '$' before the variable "
        "name (ex. $my_dict_var)."
    ),
)
@magic_arguments.argument(
    "--use_geodataframe",
    type=str,
    default=None,
    help=(
        "Return the query result as a geopandas.GeoDataFrame.  If present, the "
        "--use_geodataframe flag should be followed by a string name of the "
        "column."
    ),
)
@magic_arguments.argument(
    "--progress_bar_type",
    type=str,
    default=None,
    help=(
        "Sets progress bar type to display a progress bar while executing the query."
        "Defaults to use tqdm_notebook. Install the ``tqdm`` package to use this feature."
    ),
)
@magic_arguments.argument(
    "--location",
    type=str,
    default=None,
    help=(
        "Set the location to execute query."
        "Defaults to location set in query setting in console."
        "This flag is ignored when the engine is 'bigframes'."
    ),
)
@magic_arguments.argument(
    "--engine",
    type=str,
    default=None,
    help=(
        "Set the execution engine, either 'pandas' or 'bigframes'."
        "Defaults to engine set in the query setting in console."
    ),
)
@magic_arguments.argument(
    "--graph",
    action="store_true",
    default=False,
    help=("Visualizes the query results as a graph"),
)
@magic_arguments.argument(
    "--pyformat",
    action="store_true",
    default=False,
    help=(
        "Warning! Do not use with user-provided values. "
        "This doesn't escape values. Use --params instead for proper SQL escaping. "
        "This enables Python string formatting in the query text. "
        "Useful for values not supported by SQL query params such as table IDs. "
    ),
)
def _cell_magic(line, query):
    """Underlying function for bigquery cell magic

    Note:
        This function contains the underlying logic for the 'bigquery' cell
        magic. This function is not meant to be called directly.

    Args:
        line (str): "%%bigquery" followed by arguments as required
        query (str): SQL query to run

    Returns:
        pandas.DataFrame: the query results.
    """

    params, args = _parse_magic_args(line)

    query = query.strip()
    if not query:
        error = ValueError("Query is missing.")
        _handle_error(error, args.destination_var)
        return
    query = _validate_and_resolve_query(query, args)

    engine = args.engine or context.engine

    if engine == "bigframes":
        return _query_with_bigframes(query, params, args)

    return _query_with_pandas(query, params, args)


def _parse_magic_args(line: str) -> Tuple[List[Any], Any]:
    # The built-in parser does not recognize Python structures such as dicts, thus
    # we extract the "--params" option and interpret it separately.
    try:
        params_option_value, rest_of_args = _split_args_line(line)

    except lap.QueryParamsParseError as exc:
        raise SyntaxError(
            "--params is not a correctly formatted JSON string or a JSON "
            "serializable dictionary"
        ) from exc

    except lap.DuplicateQueryParamsError as exc:
        raise ValueError("Duplicate --params option.") from exc

    except lap.ParseError as exc:
        raise ValueError(
            "Unrecognized input, are option values correct? "
            "Error details: {}".format(exc.args[0])
        ) from exc

    params = []
    if params_option_value:
        # A non-existing params variable is not expanded and ends up in the input
        # in its raw form, e.g. "$query_params".
        if params_option_value.startswith("$"):
            msg = 'Parameter expansion failed, undefined variable "{}".'.format(
                params_option_value[1:]
            )
            raise NameError(msg)

        params = _helpers.to_query_parameters(ast.literal_eval(params_option_value), {})

    args = magic_arguments.parse_argstring(_cell_magic, rest_of_args)

    if args.engine is not None and args.engine not in ("pandas", "bigframes"):
        raise ValueError(f"Invalid engine: {args.engine}")

    return params, args


def _split_args_line(line: str) -> Tuple[str, str]:
    """Split out the --params option value from the input line arguments.

    Args:
        line: The line arguments passed to the cell magic.

    Returns:
        A tuple of two strings. The first is param option value and
        the second is the rest of the arguments.
    """
    tree = lap.Parser(lap.Lexer(line)).input_line()

    extractor = lap.QueryParamsExtractor()
    params_option_value, rest_of_args = extractor.visit(tree)

    return params_option_value, rest_of_args


def _query_with_bigframes(query: str, params: List[Any], args: Any):
    if args.dry_run:
        raise ValueError("Dry run is not supported by bigframes engine.")

    if bpd is None:
        raise ValueError("Bigframes package is not installed.")

    bpd.options.bigquery.project = context.project
    bpd.options.bigquery.credentials = context.credentials

    max_results = int(args.max_results) if args.max_results else None

    result = bpd.read_gbq_query(
        query,
        max_results=max_results,
        configuration=_create_job_config(args, params).to_api_repr(),
    )

    return _handle_result(result, args)


def _query_with_pandas(query: str, params: List[Any], args: Any):
    bq_client, bqstorage_client = _create_clients(args)
    try:
        return _make_bq_query(
            query,
            args=args,
            params=params,
            bq_client=bq_client,
            bqstorage_client=bqstorage_client,
        )
    finally:
        _close_transports(bq_client, bqstorage_client)


def _create_clients(args: Any) -> Tuple[bigquery.Client, Any]:
    bigquery_client_options = copy.deepcopy(context.bigquery_client_options)
    if args.bigquery_api_endpoint:
        if isinstance(bigquery_client_options, dict):
            bigquery_client_options["api_endpoint"] = args.bigquery_api_endpoint
        else:
            bigquery_client_options.api_endpoint = args.bigquery_api_endpoint

    bq_client = bigquery.Client(
        project=args.project or context.project,
        credentials=context.credentials,
        default_query_job_config=context.default_query_job_config,
        client_info=client_info.ClientInfo(user_agent=_get_user_agent()),
        client_options=bigquery_client_options,
        location=args.location,
    )
    if context._connection:
        bq_client._connection = context._connection

    # Check and instantiate bq storage client
    if args.use_bqstorage_api is not None:
        warnings.warn(
            "Deprecated option --use_bqstorage_api, the BigQuery "
            "Storage API is already used by default.",
            category=DeprecationWarning,
        )
    use_bqstorage_api = not args.use_rest_api and (bigquery_storage is not None)

    if not use_bqstorage_api:
        return bq_client, None

    bqstorage_client_options = copy.deepcopy(context.bqstorage_client_options)
    if args.bqstorage_api_endpoint:
        if isinstance(bqstorage_client_options, dict):
            bqstorage_client_options["api_endpoint"] = args.bqstorage_api_endpoint
        else:
            bqstorage_client_options.api_endpoint = args.bqstorage_api_endpoint

    bqstorage_client = _make_bqstorage_client(
        bq_client,
        bqstorage_client_options,
    )

    return bq_client, bqstorage_client


def _handle_result(result, args):
    """Determine the output of the cell, depending on options set.

    If an explicit destination is set, that takes precedence. Write to that
    variable and skip showing any results.

    Otherwise, if there is a default variable set (such as if this module is
    initialized by bigframes), then set that but also show the output.

    Finally, there is no variable to save to, so just show the output.
    """
    if args.destination_var:
        get_ipython().push({args.destination_var: result})
        return None

    if context.default_variable:
        # If a default variable is set, save the result _and_ show the results.
        get_ipython().push({context.default_variable: result})

    return result


def _colab_query_callback(query: str, params: str):
    return IPython.core.display.JSON(
        graph_server.convert_graph_data(query_results=json.loads(params))
    )


def _colab_node_expansion_callback(request: dict, params_str: str):
    """Handle node expansion requests in Google Colab environment

    Args:
        request: A dictionary containing node expansion details including:
            - uid: str - Unique identifier of the node to expand
            - node_labels: List[str] - Labels of the node
            - node_properties: List[Dict] - Properties of the node with key, value, and type
            - direction: str - Direction of expansion ("INCOMING" or "OUTGOING")
            - edge_label: Optional[str] - Label of edges to filter by
        params_str: A JSON string containing connection parameters

    Returns:
        JSON: A JSON-serialized response containing either:
            - The query results with nodes and edges
            - An error message if the request failed
    """
    return IPython.core.display.JSON(
        graph_server.execute_node_expansion(params_str, request)
    )


singleton_server_thread: threading.Thread = None


def _add_graph_widget(query_result):
    try:
        from spanner_graphs.graph_visualization import generate_visualization_html
    except ImportError as err:
        customized_error = ImportError(
            "Use of --graph requires the spanner-graph-notebook package to be installed. Install it with `pip install 'bigquery-magics[spanner-graph-notebook]'`."
        )
        raise customized_error from err

    # In Jupyter, create an http server to be invoked from the Javascript to populate the
    # visualizer widget. In colab, we are not able to create an http server on a
    # background thread, so we use a special colab-specific api to register a callback,
    # to be invoked from Javascript.
    port = None
    try:
        from google.colab import output

        output.register_callback("graph_visualization.Query", _colab_query_callback)
        output.register_callback(
            "graph_visualization.NodeExpansion", _colab_node_expansion_callback
        )

        # In colab mode, the Javascript doesn't use the port value we pass in, as there is no
        # graph server, but it still has to be set to avoid triggering an exception.
        # TODO: Clean this up when the Javascript is fixed on the spanner-graph-notebook side.
        port = 0
    except ImportError:
        # In this code path, we are running on Jupyter, rather than colab.
        global singleton_server_thread
        alive = singleton_server_thread and singleton_server_thread.is_alive()
        if not alive:
            singleton_server_thread = graph_server.graph_server.init()
        port = graph_server.graph_server.port

    # Create html to invoke the graph server
    html_content = generate_visualization_html(
        query="placeholder query",
        port=port,
        params=query_result.to_json().replace("\\", "\\\\").replace('"', '\\"'),
    )
    IPython.display.display(IPython.core.display.HTML(html_content))


def _is_valid_json(s: str):
    try:
        json.loads(s)
        return True
    except (json.JSONDecodeError, TypeError):
        return False


def _supports_graph_widget(query_result: pandas.DataFrame):
    # Visualization is supported if we have any json items to display.
    # (Non-json items are excluded from visualization, but we still want to bring up
    #  the visualizer for the json items.)
    for column in query_result.columns:
        if query_result[column].apply(_is_valid_json).any():
            return True
    return False


def _make_bq_query(
    query: str,
    args: Any,
    params: List[Any],
    bq_client: bigquery.Client,
    bqstorage_client: Any,
):
    max_results = int(args.max_results) if args.max_results else None
    geography_column = args.use_geodataframe

    # Any query that does not contain whitespace (aside from leading and trailing whitespace)
    # is assumed to be a table id
    if not re.search(r"\s", query):
        try:
            rows = bq_client.list_rows(query, max_results=max_results)
        except Exception as ex:
            _handle_error(ex, args.destination_var)
            return

        result = rows.to_dataframe(
            bqstorage_client=bqstorage_client,
            create_bqstorage_client=False,
        )
        return _handle_result(result, args)

    job_config = _create_job_config(args, params)
    if args.destination_table:
        split = args.destination_table.split(".")
        if len(split) != 2:
            raise ValueError(
                "--destination_table should be in a <dataset_id>.<table_id> format."
            )
        dataset_id, table_id = split
        job_config.allow_large_results = True
        dataset_ref = DatasetReference(bq_client.project, dataset_id)
        destination_table_ref = dataset_ref.table(table_id)
        job_config.destination = destination_table_ref
        job_config.create_disposition = "CREATE_IF_NEEDED"
        job_config.write_disposition = "WRITE_TRUNCATE"
        _create_dataset_if_necessary(bq_client, dataset_id)

    try:
        query_job = _run_query(bq_client, query, job_config=job_config)
    except Exception as ex:
        _handle_error(ex, args.destination_var)
        return

    if not args.verbose:
        IPython.display.clear_output()

    if args.dry_run:
        # TODO(tswast): Use _handle_result() here, too, but perhaps change the
        # format to match the dry run schema from bigframes and pandas-gbq.
        # See: https://github.com/googleapis/python-bigquery-pandas/issues/585
        if args.destination_var:
            get_ipython().push({args.destination_var: query_job})
            return
        else:
            print(
                "Query validated. This query will process {} bytes.".format(
                    query_job.total_bytes_processed
                )
            )
            return query_job

    progress_bar = context.progress_bar_type or args.progress_bar_type
    dataframe_kwargs = {
        "bqstorage_client": bqstorage_client,
        "create_bqstorage_client": False,
        "progress_bar_type": progress_bar,
    }
    if max_results:
        dataframe_kwargs["bqstorage_client"] = None

    result = query_job
    if max_results:
        result = result.result(max_results=max_results)

    if geography_column:
        result = result.to_geodataframe(
            geography_column=geography_column, **dataframe_kwargs
        )
    else:
        result = result.to_dataframe(**dataframe_kwargs)

    if args.graph and _supports_graph_widget(result):
        _add_graph_widget(result)
    return _handle_result(result, args)


def _validate_and_resolve_query(query: str, args: Any) -> str:
    # Check if query is given as a reference to a variable.
    if query.startswith("$"):
        query_var_name = query[1:]

        if not query_var_name:
            missing_msg = 'Missing query variable name, empty "$" is not allowed.'
            raise NameError(missing_msg)

        if query_var_name.isidentifier():
            ip = get_ipython()
            query = ip.user_ns.get(query_var_name, ip)  # ip serves as a sentinel

            if query is ip:
                raise NameError(
                    f"Unknown query, variable {query_var_name} does not exist."
                )
            elif not isinstance(query, (str, bytes)):
                raise TypeError(
                    f"Query variable {query_var_name} must be a string "
                    "or a bytes-like value."
                )

    if args.pyformat:
        ip = get_ipython()
        query = bigquery_magics.pyformat.pyformat(query, ip.user_ns)
    return query


def _create_job_config(args: Any, params: List[Any]) -> QueryJobConfig:
    job_config = QueryJobConfig()
    job_config.query_parameters = params
    job_config.use_legacy_sql = args.use_legacy_sql
    job_config.dry_run = args.dry_run

    # Don't override context job config unless --no_query_cache is explicitly set.
    if args.no_query_cache:
        job_config.use_query_cache = False

    if args.maximum_bytes_billed == "None":
        job_config.maximum_bytes_billed = 0
    elif args.maximum_bytes_billed is not None:
        value = int(args.maximum_bytes_billed)
        job_config.maximum_bytes_billed = value

    return job_config


def _make_bqstorage_client(client, client_options):
    """Creates a BigQuery Storage client.

    Args:
        client (:class:`~google.cloud.bigquery.client.Client`): BigQuery client.
        client_options (:class:`google.api_core.client_options.ClientOptions`):
            Custom options used with a new BigQuery Storage client instance
            if one is created.

    Raises:
        ImportError: if google-cloud-bigquery-storage is not installed, or
            grpcio package is not installed.


    Returns:
        None: if ``use_bqstorage_api == False``, or google-cloud-bigquery-storage
            is outdated.
        BigQuery Storage Client:
    """
    try:
        bigquery_magics._versions_helpers.BQ_STORAGE_VERSIONS.try_import(
            raise_if_error=True
        )
    except exceptions.BigQueryStorageNotFoundError as err:
        customized_error = ImportError(
            "The default BigQuery Storage API client cannot be used, install "
            "the missing google-cloud-bigquery-storage and pyarrow packages "
            "to use it. Alternatively, use the classic REST API by specifying "
            "the --use_rest_api magic option."
        )
        raise customized_error from err

    try:
        from google.api_core.gapic_v1 import client_info as gapic_client_info
    except ImportError as err:
        customized_error = ImportError(
            "Install the grpcio package to use the BigQuery Storage API."
        )
        raise customized_error from err

    return client._ensure_bqstorage_client(
        client_options=client_options,
        client_info=gapic_client_info.ClientInfo(user_agent=_get_user_agent()),
    )


def _close_transports(client, bqstorage_client):
    """Close the given clients' underlying transport channels.

    Closing the transport is needed to release system resources, namely open
    sockets.

    Args:
        client (:class:`~google.cloud.bigquery.client.Client`):
        bqstorage_client
            (Optional[:class:`~google.cloud.bigquery_storage.BigQueryReadClient`]):
            A client for the BigQuery Storage API.

    """
    client.close()
    if bqstorage_client is not None:
        bqstorage_client._transport.grpc_channel.close()
